/*----------------------------------------------------------------------
	This is part of the PureLS Source code.

	Copyright (C) 2000-2004 The PureLS Development Team

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
	USA
----------------------------------------------------------------------*/


/*----------------------------------------------------------------------
 Disable "unreferenced formal parameter" compiler warning
----------------------------------------------------------------------*/
#ifndef _DEBUG
	#pragma warning ( disable : 4100 )
#endif

/*----------------------------------------------------------------------
 Lets just make sure we aren't depending on IE
----------------------------------------------------------------------*/
#define _WIN32_IE 0x0000


/*---System Includes--------------------------------------------------*/
#include <tchar.h>
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
/*---Shell Includes---------------------------------------------------*/
#include "lsapi.h"
#include "../sdk/include/purels.h"
/*--------------------------------------------------------------------*/

/*---Local Definitions------------------------------------------------*/
typedef __declspec( dllimport ) INT (*FP_MODULE_INIT)(HWND, HINSTANCE, LPTSTR);
typedef __declspec( dllimport ) VOID (*FP_MODULE_QUIT)(HINSTANCE);

typedef struct _MODULEITEM
{
	HINSTANCE hInstance;

	FP_MODULE_INIT fpInit;
	FP_MODULE_QUIT fpQuit;
	//
	struct _MODULEITEM * pNext;
	//
} MODULEITEM, *PMODULEITEM;

#define MODULE_ERROR          ((UINT)-1)
#define MODULE_SUCCESS        (0x0000)
#define MODULE_INVALID_PARAM  (0x0001)
#define MODULE_LOADED         (0x0002)
#define MODULE_NOTLOADED      (0x0003)
#define MODULE_RESOURCE       (0x0004)
#define MODULE_NOTOURS        (0x0005)
#define MODULE_FAILINIT       (0x0006)
#define MODULE_NOENTRY        (0x0007)
#define MODULE_NOTFOUND       (0x0008)

#define MODULE_INIT_SUCCESS   (0)
/*--------------------------------------------------------------------*/

/*---Global Variables-------------------------------------------------*/
extern BOOL g_bIgnoreModRetVal;
extern LPTSTR g_pszAppPath;
extern HWND g_hMainWnd;

static PMODULEITEM s_pModuleList;
/*--------------------------------------------------------------------*/



static PMODULEITEM _getModuleItem( const HINSTANCE hInstance )
{
	PMODULEITEM pmi;
	
	for(pmi = s_pModuleList; NULL != pmi; pmi = pmi->pNext)
	{
		if(hInstance == pmi->hInstance)
		{
			break;
		}
	}

	return(pmi);
}

static INT _initLoadModule( const MODULEITEM * const pmi )
{
	_TCHAR szShellPath[MAX_PATH];

	/*------------------------------------------------------------------
	 Make a copy of our path, and pass it in. This way our
	 global storage can't be modified by the module.
	------------------------------------------------------------------*/
	_tcscpy(szShellPath, g_pszAppPath);

	/*------------------------------------------------------------------
	 Litestep 0.24.5 spec'd modules return:
	  Success - zero
	  Failure - non zero
	------------------------------------------------------------------*/
	return(pmi->fpInit(g_hMainWnd, pmi->hInstance, szShellPath));
}

/*--------StartModule---------------------------------------------------
 Purpose:
 - 
----------------------------------------------------------------------*/
UINT StartModule( LPCTSTR const pszModulePath )
{
	MODULEITEM mi;

	if(NULL == pszModulePath || 0 == *pszModulePath)
	{
		return(MODULE_INVALID_PARAM);
	}

	mi.hInstance = GetModuleHandle(pszModulePath);

	if(NULL != mi.hInstance)
	{
		if(NULL == _getModuleItem(mi.hInstance))
		{
			return(MODULE_NOTOURS);
		}

		return(MODULE_LOADED);
	}

	mi.hInstance = LoadLibrary(pszModulePath);

	if(NULL == mi.hInstance)
	{
		return(MODULE_RESOURCE);
	}

	mi.fpInit = (FP_MODULE_INIT)GetProcAddress(mi.hInstance, "initModuleEx" );
	mi.fpQuit = (FP_MODULE_QUIT)GetProcAddress(mi.hInstance, "quitModule" );

	if(NULL == mi.fpInit || NULL == mi.fpQuit)
	{
		FreeLibrary(mi.hInstance);
		return(MODULE_NOENTRY);
	}

	if(
		MODULE_INIT_SUCCESS == _initLoadModule(&mi)
		||
		g_bIgnoreModRetVal
	)
	{
		PMODULEITEM pmi;

		pmi = (PMODULEITEM)malloc(sizeof(MODULEITEM));

		if(NULL != pmi)
		{
			pmi->fpInit = mi.fpInit;
			pmi->fpQuit = mi.fpQuit;
			pmi->hInstance = mi.hInstance;

			pmi->pNext = s_pModuleList;
			s_pModuleList = pmi;

			return(MODULE_SUCCESS);
		}
	}

	mi.fpQuit(mi.hInstance);
	FreeLibrary(mi.hInstance);

	return(MODULE_FAILINIT);
}

static BOOL _stopModule( const HINSTANCE hInstance )
{
	PMODULEITEM pitem = s_pModuleList, pprev = NULL;

	while(NULL != pitem)
	{
		if(hInstance == pitem->hInstance)
		{
			if(NULL == pprev)
			{
				s_pModuleList = pitem->pNext;
			}
			else
			{
				pprev->pNext = pitem->pNext;
			}

			pitem->fpQuit(pitem->hInstance);
			FreeLibrary(pitem->hInstance);

			free(pitem);

			return(TRUE);
		}

		pprev = pitem;
		pitem = pitem->pNext;
	}

	return(FALSE);
}

/*--------StopModule----------------------------------------------------
 Purpose:
 - 
----------------------------------------------------------------------*/
UINT StopModule( LPCTSTR const pszModulePath )
{
	HINSTANCE hInstance;

	if(NULL == pszModulePath || 0 == *pszModulePath)
	{
		return(MODULE_INVALID_PARAM);
	}

	hInstance = GetModuleHandle(pszModulePath);

	if(NULL == hInstance)
	{
		return(MODULE_NOTLOADED);
	}

	if(!_stopModule(hInstance))
	{
		return(MODULE_NOTOURS);
	}

	CleanDeadMessages();
	CleanDeadDialogs();

	return(MODULE_SUCCESS);
}

/*--------ReloadModule--------------------------------------------------
 Purpose:
 - 
----------------------------------------------------------------------*/
UINT ReloadModule( LPCTSTR const pszModulePath )
{
	UINT uRetVal;

	uRetVal = StopModule(pszModulePath);

	if(MODULE_SUCCESS == uRetVal)
	{
		uRetVal = StartModule(pszModulePath);
	}

	return(uRetVal);
}

/*--------AddModules----------------------------------------------------
 Purpose:
 - 
----------------------------------------------------------------------*/
BOOL AddModules( VOID )
{
	_TCHAR szBuffer[MAX_PATH];
	UINT uIndex = 0;

	while(SH_CFG_GetLoadModule(&uIndex, szBuffer))
	{
		if(MODULE_SUCCESS != StartModule(szBuffer))
		{
			WarningMessageBox(szBuffer, _T("PureLS - Error loading module:"), MB_OK);
		}
	}

	return(NULL != s_pModuleList);
}

/*--------RemoveModules-------------------------------------------------
 Purpose:
 - 
----------------------------------------------------------------------*/
VOID RemoveModules( VOID )
{
	PMODULEITEM pmi;

	while(NULL != s_pModuleList)
	{
		pmi = s_pModuleList->pNext;

		s_pModuleList->fpQuit(s_pModuleList->hInstance);
		FreeLibrary(s_pModuleList->hInstance);
		free(s_pModuleList);

		s_pModuleList = pmi;
	}

	return;
}

/*--------End Of File-------------------------------------------------*/
