/*----------------------------------------------------------------------
	This is part of the PureLS Source code.

	Copyright (C) 2000-2003 The PureLS Development Team

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
	USA
----------------------------------------------------------------------*/


/*
 EPSHeader

   File: match.c
   Author: J. Kercheval
   Created: Sat, 01/05/1991  22:21:49
*/
/*
 EPSRevision History

   J. Kercheval  Wed, 02/20/1991  22:29:01  Released to Public Domain
   J. Kercheval  Fri, 02/22/1991  15:29:01  fix '\' bugs (two :( of them)
   J. Kercheval  Sun, 03/10/1991  19:31:29  add error return to matche()
   J. Kercheval  Sun, 03/10/1991  20:11:11  add is_valid_pattern code
   J. Kercheval  Sun, 03/10/1991  20:37:11  beef up main()
   J. Kercheval  Tue, 03/12/1991  22:25:10  Released as V1.1 to Public Domain
*/

/*
 PureLS Revision History

  - Changed all character comparison to TCHAR defines to support UNICODE builds
*/

/*
   Wildcard Pattern Matching
*/
#include <tchar.h>
#include <windows.h>
#include <stdio.h>
/* MinGW doesn't include isspace() and so automatically */
#include <ctype.h>

/*** Shell Includes ************/
#include "lsapi.h"
#include "../sdk/include/purels.h"
/*******************************/

#define MATCH_PATTERN  6    /* bad pattern */
#define MATCH_LITERAL  5    /* match failure on literal match */
#define MATCH_RANGE    4    /* match failure on [..] construct */
#define MATCH_ABORT    3    /* premature end of text string */
#define MATCH_END      2    /* premature end of pattern string */
#define MATCH_VALID    1    /* valid match */

/* pattern defines */
#define PATTERN_VALID  0    /* valid pattern */
#define PATTERN_ESC   -1    /* literal escape at end of pattern */
#define PATTERN_RANGE -2    /* malformed range in [..] construct */
#define PATTERN_CLOSE -3    /* no end bracket in [..] construct */
#define PATTERN_EMPTY -4    /* [..] contstruct is empty */


INT matche_after_star (LPCTSTR pattern, LPCTSTR text);

/*----------------------------------------------------------------------------
*
* Return TRUE if PATTERN has any special wildcard characters
*
----------------------------------------------------------------------------*/
BOOL is_pattern (LPCTSTR p)
{
	while (*p)
	{
		switch (*p++)
		{
		case _T('?'):
		case _T('*'):
		case _T('['):
		case _T('\\'):
			return TRUE;
		}
	}
	return FALSE;
}


/*----------------------------------------------------------------------------
*
* Return TRUE if PATTERN has is a well formed regular expression according
* to the above syntax
*
* error_type is a return code based on the type of pattern error.  Zero is
* returned in error_type if the pattern is a valid one.  error_type return
* values are as follows:
*
*   PATTERN_VALID - pattern is well formed
*   PATTERN_ESC   - pattern has invalid escape ('\' at end of pattern)
*   PATTERN_RANGE - [..] construct has a no end range in a '-' pair (ie [a-])
*   PATTERN_CLOSE - [..] construct has no end bracket (ie [abc-g )
*   PATTERN_EMPTY - [..] construct is empty (ie [])
*
----------------------------------------------------------------------------*/
BOOL is_valid_pattern (LPCTSTR p, PINT error_type)
{
	/* init error_type */
	*error_type = PATTERN_VALID;
	
	/* loop through pattern to EOS */
	while (0 != *p)
	{
		/* determine pattern type */
		switch (*p)
		{
        /* check literal escape, it cannot be at end of pattern */
		case _T('\\'):
			if (0 == *++p)
			{
				*error_type = PATTERN_ESC;
				return FALSE;
			}
			p++;
			break;
			
		/* the [..] construct must be well formed */
		case _T('['):
			p++;
			
			/* if the next character is ']' then bad pattern */
			if (_T(']') == *p)
			{
				*error_type = PATTERN_EMPTY;
				return FALSE;
			}
			
			/* if end of pattern here then bad pattern */
			if (0 == *p)
			{
				*error_type = PATTERN_CLOSE;
				return FALSE;
			}
			
			/* loop to end of [..] construct */
			while (_T(']') != *p)
			{
				/* check for literal escape */
				if (_T('\\') == *p++)
				{
					if (0 == *p++)
					{
						*error_type = PATTERN_ESC;
						return FALSE;
					}
				}

				/* if end of pattern here then bad pattern */
				if (0 == *p)
				{
					*error_type = PATTERN_CLOSE;
					return FALSE;
				}
				
				/* if this a range */
				if (*p == _T('-'))
				{
					/* we must have an end of range */
					if (0 == *++p || _T(']') == *p)
					{
						*error_type = PATTERN_RANGE;
						return FALSE;
					}
				}
			}
			break;
			
		/* all other characters are valid pattern elements */
		case _T('*'):
		case _T('?'):
		default:
			p++;                              /* "normal" character */
			break;
		}
	}
	return TRUE;
}


/*----------------------------------------------------------------------------
*
*  Match the pattern PATTERN against the string TEXT;
*
*  returns MATCH_VALID if pattern matches, or an errorcode as follows
*  otherwise:
*
*            MATCH_PATTERN  - bad pattern
*            MATCH_LITERAL  - match failure on literal mismatch
*            MATCH_RANGE    - match failure on [..] construct
*            MATCH_ABORT    - premature end of text string
*            MATCH_END      - premature end of pattern string
*            MATCH_VALID    - valid match
*
*
*  A match means the entire string TEXT is used up in matching.
*
*  In the pattern string:
*       `*' matches any sequence of characters (zero or more)
*       `?' matches any character
*       [SET] matches any character in the specified set,
*       [!SET] or [^SET] matches any character not in the specified set.
*
*  A set is composed of characters or ranges; a range looks like
*  character hyphen character (as in 0-9 or A-Z).  [0-9a-zA-Z_] is the
*  minimal set of characters allowed in the [..] pattern construct.
*  Other characters are allowed (ie. 8 bit characters) if your system
*  will support them.
*
*  To suppress the special syntactic significance of any of `[]*?!^-\',
*  and match the character exactly, precede it with a `\'.
*
----------------------------------------------------------------------------*/
INT matche (LPCTSTR p, LPCTSTR t)
{
	_TCHAR range_start, range_end;  /* start and end in range */
	
	BOOL invert;             /* is this [..] or [!..] */
	BOOL member_match;       /* have I matched the [..] construct? */
	BOOL loop;               /* should I terminate? */
	
	for ( ; 0 != *p; p++, t++)
	{
		/* If this is the end of the text then this is the end of the match */
		if (0 == *t)
		{
			return (_T('*') == *p && 0 == *++p ) ? MATCH_VALID : MATCH_ABORT;
		}
		
		/* determine and react to pattern type */
		switch (*p)
		{
		/* single any character match */
		case _T('?'):
			break;
			
		/* multiple any character match */
		case _T('*'):
			return matche_after_star(p, t);
			
		/* [..] construct, single member/exclusion character match */
		case _T('['):
			/* move to beginning of range */
			p++;

			/* check if this is a member match or exclusion match */
			if (_T('!') == *p || _T('^') == *p)
			{
				invert = TRUE;
				p++;
			}
			else
			{
				invert = FALSE;
			}
			
			/* if closing bracket here or at range start then we have a
			 malformed pattern */
			if (_T(']') == *p)
			{
				return MATCH_PATTERN;
			}
			
			member_match = FALSE;
			loop = TRUE;
			
			while (loop)
			{
				/* if end of construct then loop is done */
				if (_T(']') == *p)
				{
					loop = FALSE;
					continue;
				}
				
				/* matching a '!', '^', '-', '\' or a ']' */
				if (_T('\\') == *p)
				{
					p++;
				}

				range_start = range_end = *p++;
				
				/* if end of pattern then bad pattern (Missing ']') */
				if (0 == range_start)
				{
					return MATCH_PATTERN;
				}
				
				/* check for range bar */
				if (_T('-') == *p)
				{
					/* get the range end */
					range_end = *++p;
					
					/* if end of pattern or construct then bad pattern */
					if (0 == range_end || _T(']') == range_end)
					{
						return MATCH_PATTERN;
					}
					
					/* special character range end */
					if (_T('\\') == range_end)
					{
						range_end = *++p;
						
						/* if end of text then we have a bad pattern */
						if (0 == range_end)
						{
							return MATCH_PATTERN;
						}
					}
					
					/* move just beyond this range */
					p++;
				}
				
				/* if the text character is in range then match found.
				 make sure the range letters have the proper
				 relationship to one another before comparison */
				if (range_start < range_end)
				{
					if (*t >= range_start && *t <= range_end)
					{
						member_match = TRUE;
						loop = FALSE;
					}
				}
				else
				{
					if (*t >= range_end && *t <= range_start)
					{
						member_match = TRUE;
						loop = FALSE;
					}
				}
			}
			
			/* if there was a match in an exclusion set then no match */
			/* if there was no match in a member set then no match */
			if ((invert && member_match) || !(invert || member_match))
			{
				return MATCH_RANGE;
			}
			
			/* if this is not an exclusion then skip the rest of
			 the [...] construct that already matched. */
			if (member_match)
			{
				while (_T(']') != *p)
				{
					/* bad pattern (Missing ']') */
					if (0 == *p)
					{
						return MATCH_PATTERN;
					}

					/* skip exact match */
					if (_T('\\') == *p++)
					{
						/* if end of text then we have a bad pattern */
						if (0 == *p++)
						{
							return MATCH_PATTERN;
						}
					}
				}
			}
			break;

		/* next character is quoted and must match exactly */
		case _T('\\'):

			/* move pattern pointer to quoted char and fall through */
			p++;
			
			/* if end of text then we have a bad pattern */
			if (0 == *p)
			{
				return MATCH_PATTERN;
			}
			
			/* must match this character exactly */
			/* FALL THROUGH */
			
		default:
			if (toupper(*p) != toupper(*t))
			{
				return MATCH_LITERAL;
			}
		}
	}

	/* if end of text not reached then the pattern fails */
	if (0 != *t)
	{
		return MATCH_END;
	}

	return MATCH_VALID;
}


/*----------------------------------------------------------------------------
*
* recursively call matche() with final segment of PATTERN and of TEXT.
*
----------------------------------------------------------------------------*/
int matche_after_star (LPCTSTR p, LPCTSTR t)
{
	INT match = 0;
	_TCHAR nextp;

	/* pass over existing ? and * in pattern */
	while (_T('?') == *p || _T('*') == *p)
	{
		/* take one char for each ? and + */
		if (_T('?') == *p++)
		{
			/* if end of text then no match */
			if (0 == *t++)
			{
				return MATCH_ABORT;
			}
		}
	}
	
	/* if end of pattern we have matched regardless of text left */
	if (0 == *p)
	{
		return MATCH_VALID;
	}

	/* get the next character to match which must be a literal or '[' */
	nextp = *p;

	if (_T('\\') == nextp)
	{
		nextp = p[1];
		
		/* if end of text then we have a bad pattern */
		if (0 == nextp)
		{
			return MATCH_PATTERN;
		}
	}
	
	/* Continue until we run out of text or definite result seen */
	do
	{
		/* a precondition for matching is that the next character
		 in the pattern match the next character in the text or that
		 the next pattern char is the beginning of a range.  Increment
		 text pointer as we go here */
		if (tolower(nextp) == tolower(*t) || _T('[') == nextp)
		{
			match = matche(p, t);
		}
		
		/* if the end of text is reached then no match */
		if (!*t++)
		{
			match = MATCH_ABORT;
		}
	} while (match != MATCH_VALID && match != MATCH_ABORT && match != MATCH_PATTERN);
	
	/* return result */
	return match;
}


/*----------------------------------------------------------------------------
*
* match() is a shell to matche() to return only BOOL values.
*
----------------------------------------------------------------------------*/

BOOL match( LPCTSTR p, LPCTSTR t )
{
	INT error_type;

	error_type = matche(p,t);

	return MATCH_VALID == error_type;
}
